import requests
from requests.exceptions import ConnectionError, ConnectTimeout

from . import settings

OK_STATUS = {'status': 'ok'}


def health():
    url = f'{settings.get_url()}/api/health/'

    try:
        response = requests.get(url, timeout=settings.get_timeout())
    except (ConnectTimeout, ConnectionError):
        return False

    return response.json() == OK_STATUS



from django.core.management.base import BaseCommand

from ...config import configure_topics


class Command(BaseCommand):
    help = 'Create all topics to be able to publish messages to postoffice'

    def handle(self, *args, **kwargs):
        configure_topics()



class MessagesSerializer:
    fields = ['id', 'topic', 'payload', 'attributes', 'bulk']

    def serialize(self, messages):
        return [self._serialize_fields(message) for message in messages]

    def _serialize_fields(self, message):
        return {field: getattr(message, field, None) for field in self.fields}



import logging
from collections import namedtuple
from http import HTTPStatus

import requests
from django.urls import reverse
from requests.exceptions import ConnectionError, ConnectTimeout

from postoffice_django.exceptions import BadPublisherCreation, BadTopicCreation

from . import settings

logger = logging.getLogger(__name__)

ConfigurationResponse = namedtuple('ConfigurationResponse', ['report_error'])


def configure_publishers() -> None:
    uncreated_publishers = []

    for consumer in settings.get_consumers():
        if _create_publishers(consumer).report_error:
            uncreated_publishers.append(consumer)

    if uncreated_publishers:
        raise BadPublisherCreation(uncreated_publishers)


def configure_topics() -> None:
    uncreated_topics = []

    for topic in settings.get_topics():
        if _create_topic(topic).report_error:
            uncreated_topics.append(topic)

    if uncreated_topics:
        raise BadTopicCreation(uncreated_topics)


def _create_publishers(consumer: dict) -> ConfigurationResponse:
    url = f'{settings.get_url()}/api/publishers/'
    payload = {
        'active': True,
        'topic': consumer.get('topic'),
        'target': consumer.get('target'),
        'type': consumer.get('type'),
        'from_now': True
    }

    payload.update(_get_optional_args(consumer))

    return _execute_request(url, payload)


def _get_optional_args(consumer: dict) -> dict:
    return {**_get_publisher_retry(consumer), **_get_publisher_timeout(consumer)}


def _get_publisher_timeout(consumer: dict) -> dict:
    if not consumer.get('timeout'):
        return {}
    return {'seconds_timeout': consumer.get('timeout')}


def _get_publisher_retry(consumer: dict) -> dict:
    if not consumer.get('retry'):
        return {}
    return {'seconds_retry': consumer.get('retry')}


def _create_topic(topic_name: str) -> ConfigurationResponse:
    url = f'{settings.get_url()}/api/topics/'
    payload = {'name': topic_name, 'origin_host': _generate_origin_host()}

    return _execute_request(url, payload)


def _generate_origin_host() -> str:
    return settings.get_origin_host() + reverse('postoffice-messages-list')


def _execute_request(url: str, payload: dict) -> ConfigurationResponse:
    try:
        response = requests.post(
            url,
            json=payload,
            timeout=settings.get_timeout()
        )
    except (ConnectTimeout, ConnectionError):
        return ConfigurationResponse(report_error=True)

    if response.status_code == HTTPStatus.CONFLICT:
        logger.warning('Existing resource',
                       extra={'url': url, 'payload': payload})
        return ConfigurationResponse(report_error=False)

    if response.status_code == HTTPStatus.CREATED:
        return ConfigurationResponse(report_error=False)

    return ConfigurationResponse(report_error=True)



import json
from datetime import timedelta

import requests
from django.core.serializers.json import DjangoJSONEncoder
from django.utils import timezone
from requests import Response
from requests.exceptions import ConnectionError, Timeout

from . import settings
from .models import PublishingError

CONNECTION_ERROR = 'Can not establish connection with postoffice'


def publish(topic: str, payload: dict, **attrs: dict) -> None:
    Publisher(topic, payload, **attrs).publish()


def bulk_publish(topic: str, payload: list, **attrs: dict) -> None:
    BulkPublisher(topic, payload, **attrs).publish()


def scheduled_publish(topic: str, payload: dict, schedule_in: int, **attrs: dict) -> None:
    ScheduledPublisher(topic, payload, schedule_in, **attrs).publish()


class Publisher:
    URL = 'api/messages/'
    TIMEOUT = settings.get_timeout()

    def __init__(self, topic, payload, bulk=False, **attributes):
        self.url = f'{settings.get_url()}/{self.URL}'
        self.topic = topic
        self.payload = payload
        self.attributes = self._stringify(attributes)
        self.timeout = self.TIMEOUT
        self.bulk = bulk
        self.message = self._create_message()

    def publish(self):
        try:
            message = json.dumps(self.message, cls=DjangoJSONEncoder)
            response = requests.post(
                self.url, data=message, headers={'Content-Type': 'application/json'}, timeout=self.timeout)
        except (ConnectionError, Timeout):
            self._save_connection_not_established()
            return

        if response.status_code != 201:
            self._save_publishing_error(response)

    def _create_message(self) -> dict:
        return {
            'topic': self.topic,
            'payload': self.payload,
            'attributes': self.attributes,
        }

    def _stringify(self, attributes) -> dict:
        return {key: str(attributes[key]) for key in attributes.keys()}

    def _save_connection_not_established(self) -> None:
        self._create_publishing_error(CONNECTION_ERROR)

    def _save_publishing_error(self, response: Response) -> None:
        error = 'Unknown error'

        if response.status_code == 500:
            error = 'Internal server error'

        if response.status_code == 400:
            error = response.json().get('data').get('errors')

        self._create_publishing_error(error)

    def _create_publishing_error(self, error: str) -> None:
        PublishingError.objects.create(
            topic=self.topic,
            payload=self._error_payload(),
            attributes=self.attributes,
            bulk=self.bulk,
            error=error,
        )

    def _error_payload(self):
        return self.payload


class BulkPublisher(Publisher):
    URL = 'api/bulk_messages/'
    TIMEOUT = settings.get_bulk_timeout()

    def __init__(self, topic, payload, **attributes):
        super().__init__(topic=topic, payload=payload, bulk=True, **attributes)

    def _create_publishing_error(self, error: str) -> None:
        PublishingError.objects.create(
            topic=self.topic,
            payload=self._create_failed_message(),
            attributes=self.attributes,
            bulk=self.bulk,
            error=error,
        )

    def _create_failed_message(self) -> list:
        return [{
            'topic': self.topic,
            'attributes': self.attributes,
            'payload': message_payload
        } for message_payload in self.payload]

    def _error_payload(self):
        return self.message


class ScheduledPublisher(Publisher):
    URL = 'api/schedule_messages/'

    def __init__(self, topic, payload, schedule_in, **attributes):
        self.schedule_in = schedule_in
        super().__init__(topic=topic, payload=payload, **attributes)

    def _create_message(self) -> dict:
        message = super()._create_message()
        message['schedule_at'] = timezone.now() + timedelta(minutes=self.schedule_in)
        return message



class UrlSettingNotDefined(Exception):
    def __init__(self):
        self.message = 'POSTOFFICE_URL is not defined on settings'


class OriginHostSettingNotDefined(Exception):
    def __init__(self):
        self.message = 'ORIGIN_HOST is not defined on settings'


class BadTopicCreation(Exception):
    def __init__(self, topic_name):
        self.message = (
            f'Can not create topic. Topic no created: { topic_name }')


class BadPublisherCreation(Exception):
    def __init__(self, publisher):
        self.message = (
            f'Can not create publisher. Publisher not created: { publisher }')



import logging
from typing import Any, List, Union

from django.conf import settings

from .exceptions import (
    OriginHostSettingNotDefined,
    UrlSettingNotDefined
)

logger = logging.getLogger(__name__)

DEFAULT_TIMEOUT = 0.5
DEFAULT_BULK_TIMEOUT = 5


def get_url() -> str:
    try:
        return settings.POSTOFFICE['URL']
    except KeyError:
        raise UrlSettingNotDefined


def get_consumers() -> List[dict]:
    try:
        return settings.POSTOFFICE['CONSUMERS']
    except KeyError:
        logger.warning('Consumers config key is missing')
        return []


def get_timeout() -> Union[float, Any]:
    try:
        timeout = settings.POSTOFFICE['TIMEOUT']
        return float(timeout)
    except KeyError:
        logger.info(
            f'Timeout not defined, using default value: {DEFAULT_TIMEOUT} (s)')
        return DEFAULT_TIMEOUT


def get_bulk_timeout() -> Union[float, Any]:
    try:
        timeout = settings.POSTOFFICE['BULK_TIMEOUT']
        return float(timeout)
    except KeyError:
        logger.info(
            'Bulk timeout not defined, '
            f'using default: {DEFAULT_BULK_TIMEOUT} (s)')
        return DEFAULT_BULK_TIMEOUT


def get_origin_host() -> str:
    try:
        return settings.POSTOFFICE['ORIGIN_HOST']
    except KeyError:
        raise OriginHostSettingNotDefined


def get_topics() -> List[str]:
    try:
        return settings.POSTOFFICE['TOPICS']
    except KeyError:
        logger.warning('Topics config key is missing')
        return []



import json
from unittest.mock import patch

import pytest

from postoffice_django.models import PublishingError


@pytest.mark.django_db
class TestListMessagesView:

    def test_returns_empty_when_no_publishing_errors_exist(self, client):
        response = client.get('/messages/')

        assert response.status_code == 200
        assert json.loads(response.content) == []

    def test_returns_ordered_messages_when_publishing_errors_exist(
            self, client, publishing_error, older_publishing_error):
        response = client.get('/messages/')

        assert response.status_code == 200
        assert json.loads(response.content) == [{
            'id': older_publishing_error.id,
            'topic': 'test-topic',
            'attributes': None,
            'payload': {'approved': True},
            'bulk': False,
        }, {
            'id': publishing_error.id,
            'topic': 'test',
            'payload': {'key': 'value', 'num': '2.15', 'elements': [1, 2, 3]},
            'attributes': {'key': 'value'},
            'bulk': False,
        }]

    def test_returns_bulk_message_when_bulk_publishing_error_exists(
            self, client, bulk_publishing_error):
        response = client.get('/messages/')

        assert response.status_code == 200
        assert json.loads(response.content) == [{
            'id': bulk_publishing_error.id,
            'topic': 'test-topic',
            'attributes': None,
            'payload': [
                {'topic': 'test-topic', 'payload': {'approved': True}},
                {'topic': 'test-topic', 'payload': {'approved': False}},
            ],
            'bulk': True,
        }]

    @patch(
        'postoffice_django.api.views.ListMessagesView.DEFAULT_MAX_RESULTS', 1)
    def test_only_default_messages_quantity_returned_when_no_limit_received(
            self, client, publishing_error, older_publishing_error):
        response = client.get('/messages/')

        assert response.status_code == 200
        assert len(json.loads(response.content)) == 1

    @patch(
        'postoffice_django.api.views.ListMessagesView.DEFAULT_MAX_RESULTS', 1)
    def test_all_publishing_errors_returned_when_bigger_limit_received(
            self, client, publishing_error, older_publishing_error):
        response = client.get('/messages/?limit=10')

        assert response.status_code == 200
        publishing_errors_count = PublishingError.objects.count()
        assert len(json.loads(response.content)) == publishing_errors_count


@pytest.mark.django_db
class TestDeleteMessageView:

    def test_returns_no_content_when_message_is_deleted(
            self, client, publishing_error):
        response = client.delete(f'/messages/{publishing_error.id}/')

        assert response.status_code == 204
        assert response.content == b''

        assert PublishingError.objects.count() == 0

    def test_only_received_message_is_deleted_when_multiple_messages_exist(
            self, client, publishing_error, older_publishing_error):
        client.delete(f'/messages/{publishing_error.id}/')

        assert PublishingError.objects.count() == 1
        assert PublishingError.objects.first() == older_publishing_error

    def test_returns_not_found_when_id_does_not_exist(
            self, client, publishing_error):
        response = client.delete('/messages/0/')

        assert response.status_code == 404



from django.contrib.postgres.fields import JSONField
from django.db import models


class PublishingError(models.Model):
    topic = models.CharField(max_length=100)
    payload = JSONField()
    attributes = JSONField(null=True)
    bulk = models.BooleanField(default=False)
    error = models.CharField(max_length=255, default='')
    created_at = models.DateTimeField(auto_now_add=True)



from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('postoffice_django', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='publishingerror',
            name='bulk',
            field=models.BooleanField(default=False),
        ),
    ]



from django.http import JsonResponse, HttpResponse
from django.views import View
from django.utils.decorators import method_decorator
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.edit import BaseDeleteView

from postoffice_django.models import PublishingError
from postoffice_django.serializers import MessagesSerializer


class ListMessagesView(View):
    DEFAULT_MAX_RESULTS = 100

    def get(self, request, *args, **kwargs):
        max_results = self._get_max_results(request)
        messages = PublishingError.objects.order_by('created_at')[:max_results]
        data = MessagesSerializer().serialize(messages)
        return JsonResponse(data, safe=False)

    def _get_max_results(self, request):
        return int(request.GET.get('limit', self.DEFAULT_MAX_RESULTS))


@method_decorator(csrf_exempt, name='dispatch')
class DeleteMessageView(BaseDeleteView):
    queryset = PublishingError.objects.all()

    def delete(self, request, *args, **kwargs):
        message = self.get_object()
        message.delete()
        return HttpResponse(status=204)



from django.urls import path

from postoffice_django.api.views import DeleteMessageView, ListMessagesView

urlpatterns = [
    path(
        'messages/',
        ListMessagesView.as_view(),
        name='postoffice-messages-list'
    ),
    path(
        'messages/<int:pk>/',
        DeleteMessageView.as_view(),
        name='postoffice-messages-detail'
    ),
]



import os
import sys
from os import path

from setuptools import find_packages, setup
from setuptools.command.install import install

README = open(os.path.join(os.path.dirname(__file__), 'README.md')).read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

current_directory = path.abspath(path.dirname(__file__))

with open(path.join(current_directory, 'README.md'), encoding='utf-8') as file:
    long_description = file.read()


VERSION = '0.8.0'


class VerifyVersionCommand(install):
    description = 'Verify that the git tag matches our version'

    def run(self):
        tag = os.getenv('CIRCLE_TAG')

        if tag != VERSION:
            info = (f'Git tag: {tag} does not match the '
                    f'version of this app: {VERSION}')
            sys.exit(info)


setup(
    name='postoffice_django',
    version=VERSION,
    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    license='APACHE License',
    description='A simple Django app to comunicate with postoffice',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/mercadona/postoffice_django/',
    author='Mercadona',
    author_email='sofware.online@mercadona.es',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django :: 2.0',
        'Framework :: Django :: 2.1',
        'Framework :: Django :: 2.2',
        'Framework :: Django :: 3.0',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    cmdclass={
        'verify': VerifyVersionCommand,
    }
)
